#
# File:    ./src/vutils/yaml/utils.py
# Author:  Jiří Kučera <sanczes AT gmail.com>
# Date:    2022-06-23 21:40:39 +0200
# Project: vutils-yaml: Working with YAML format
#
# SPDX-License-Identifier: MIT
#
"""Miscellaneous utilities."""

import datetime
from typing import TYPE_CHECKING, Iterator, cast

import yaml
from vutils.python.data import merge_data
from vutils.validator.value import Location

if TYPE_CHECKING:
    from collections.abc import Callable

    from vutils.yaml import (
        PyDict,
        PyList,
        PySet,
        _CtorDecorType,
        _CtorFuncType,
        _CtorSpecType,
        _CtorType,
        _MarkType,
        _new_date,
        _new_datetime,
        _NodeType,
    )
else:
    PyList = list
    PySet = set
    PyDict = dict
    _new_date = datetime.date.__new__
    _new_datetime = datetime.datetime.__new__

ANNOTATION_SLOT: str = "__yaml_annotation__"
KEYLOC_SLOT: str = "__yaml_keyloc__"


class Annotation:
    """Holds YAML values annotation."""

    __slots__ = ("location",)

    def __init__(self, location: "Location | None" = None) -> None:
        """
        Initialize annotation object.

        :param location: The YAML value's location
        """
        if location is None:
            location = Location()
        self.location = location


class YamlDataType:
    """Base class for annotated YAML objects."""

    __slots__ = ()


class NullType(YamlDataType):
    """
    A null (`None`) type.

    Since `None` has a special meaning it cannot be annotated nor it is
    possible to inherit from its type. This workaround is used to store
    annotation alongside with ``null``.
    """

    def __bool__(self) -> bool:
        """
        Convert `NullType` object to the `bool` object.

        :return: always `False`
        """
        return False

    def __eq__(self, other: object) -> bool:
        """
        Test the equality of this and the other object.

        :return: `True` if this and the other object are considered equal
        """
        return other is None or isinstance(other, type(self))

    def __hash__(self) -> int:
        """
        Return the object hash.

        :return: the hash of this object
        """
        return hash(None)


class BoolType(YamlDataType):
    """
    Represent a Boolean type.

    Since `bool` is not an acceptable base class this workaround is used to
    store annotation alongside with Boolean values.
    """

    def __init__(self, value: bool) -> None:
        """
        Initialize the `bool`-like object.

        :param value: The value
        """
        YamlDataType.__init__(self)
        self.__value = value

    def __bool__(self) -> bool:
        """
        Convert this object to `bool` object.

        :return: the kept value
        """
        return self.__value

    def __eq__(self, other: object) -> bool:
        """
        Test the equality of this and the other object.

        :return: `True` if this and the other objects are considered equal
        """
        if isinstance(other, (bool, BoolType)):
            return bool(self) is bool(other)
        if isinstance(other, (int, float)):
            return int(bool(self)) == other
        return False

    def __hash__(self) -> int:
        """
        Return the object hash.

        :return: the hash of this object
        """
        return hash(bool(self))


class IntType(int, YamlDataType):
    """
    Wrap the `int` type.

    Needed to store annotation.
    """


class FloatType(float, YamlDataType):
    """
    Wrap the `float` type.

    Needed to store annotation.
    """


class StrType(str, YamlDataType):
    """
    Wrap the `str` type.

    Needed to store annotation.
    """


class BytesType(bytes, YamlDataType):
    """
    Wrap the `bytes` type.

    Needed to store annotation.
    """


class ListType(PyList, YamlDataType):
    """
    Wrap the `list` type.

    Needed to store annotation.
    """


class SetType(PySet, YamlDataType):
    """
    Wrap the `set` type.

    Needed to store annotation.
    """


class DictType(PyDict, YamlDataType):
    """
    Wrap the `dict` type.

    Needed to store annotation.
    """


class DateType(datetime.date, YamlDataType):
    """
    Wrap the `datetime.date` type.

    Needed to store annotation.
    """

    def __new__(cls, *args: object) -> "DateType":
        """
        Create the date object.

        :param args: Either `datetime.date` object or `datetime.date` arguments
        :return: the date object
        """
        if isinstance(args[0], datetime.date):
            date: datetime.date = args[0]
            return datetime.date.__new__(cls, date.year, date.month, date.day)
        return _new_date(cls, *args)


class DateTimeType(datetime.datetime, YamlDataType):
    """
    Wrap the `datetime.datetime` type.

    Needed to store annotation.
    """

    def __new__(cls, *args: object, **kwargs: object) -> "DateTimeType":
        """
        Create the datetime object.

        :param args: Either `datetime.datetime` object or `datetime.datetime`
            arguments
        :param kwargs: Key-word arguments to `datetime.datetime`
        :return: the datetime object
        """
        if isinstance(args[0], datetime.datetime):
            stamp: datetime.datetime = args[0]
            return datetime.datetime.__new__(
                cls,
                stamp.year,
                stamp.month,
                stamp.day,
                stamp.hour,
                stamp.minute,
                stamp.second,
                stamp.microsecond,
                stamp.tzinfo,
                fold=stamp.fold,
            )
        return _new_datetime(cls, *args, **kwargs)


TYPEMAP: "dict[object, type[YamlDataType]]" = {
    bool: BoolType,
    int: IntType,
    float: FloatType,
    str: StrType,
    bytes: BytesType,
    list: ListType,
    set: SetType,
    dict: DictType,
    datetime.date: DateType,
    datetime.datetime: DateTimeType,
}


def obj2xobj(obj: object) -> YamlDataType:
    """
    Make object extensible.

    :param obj: The object
    :return: the extensible object
    :raises TypeError: when object cannot be made extensible

    Make object extensible so additional information can be added to it. This
    changes the object type.
    """
    if obj is None:
        return NullType()
    cls: "type[YamlDataType] | None" = TYPEMAP.get(type(obj))
    if cls is None:
        raise TypeError(f"{type(obj)} object cannot be made extensible")
    return cast("Callable[[object], YamlDataType]", cls)(obj)


def newc(name: str, args: "tuple[object, ...]") -> object:
    """
    Create a new container based on the constructor name.

    :param name: The constructor name
    :param args: The constructor arguments
    :return: the new container
    """
    if name == "construct_yaml_set":
        return set()
    if name == "construct_yaml_map":
        return {}
    if name == "construct_yaml_object":
        return cast("Callable[[object], object]", args[0].__new__)(args[0])
    return []


def annotate(data: object, node: "_NodeType") -> YamlDataType:
    """
    Annotate data with information from node.

    :param data: The data object
    :param node: The node
    :return: the annotated data object
    :raises yaml.error.YAMLError: when operation fails
    """
    mark: "_MarkType" = node.start_mark
    name: str = mark.name
    line: int = mark.line + 1
    column: int = mark.column + 1

    ydata: YamlDataType = cast(YamlDataType, data)
    if not hasattr(data, ANNOTATION_SLOT):
        try:
            ydata = obj2xobj(data)
        except TypeError as exc:
            detail: str = cast("tuple[str]", exc.args)[0]
            raise yaml.constructor.ConstructorError(
                f"{name}:{line}:{column}: {detail}"
            )
        setattr(ydata, ANNOTATION_SLOT, Annotation())

    location: Location = getloc(ydata)
    location.path = name
    location.line = line
    location.column = column

    return ydata


def make_ctor(name: str, base_cls: "type[_CtorType]") -> "_CtorFuncType":
    """
    Create YAML object constructor (function).

    :param name: The constructor function name
    :param base_cls: The base of the constructor class
    :return: the constructor function
    """

    def ctor(
        inst: "_CtorType", node: "_NodeType", *args: object, **kwargs: object
    ) -> object:
        """
        Construct an annotated YAML object.

        :param inst: The YAML constructor class instance
        :param node: The node
        :param args: Additional arguments
        :param kwargs: Additional keyword arguments
        :return: the annotated YAML object
        """
        base_ctor: "_CtorFuncType" = cast(
            "_CtorFuncType", getattr(base_cls, name)
        )
        data: object = base_ctor(inst, node, *args, **kwargs)
        return annotate(data, node)

    return ctor


def make_gctor(name: str, base_cls: "type[_CtorType]") -> "_CtorFuncType":
    """
    Create YAML object constructor (generator).

    :param name: The constructor name
    :param base_cls: The base of the constructor class
    :return: the constructor generator
    """

    def gctor(
        inst: "_CtorType", node: "_NodeType", *args: object, **kwargs: object
    ) -> Iterator[object]:
        """
        Generate an annotated YAML object.

        :param inst: The YAML constructor class instance
        :param node: The node
        :param args: Additional arguments
        :param kwargs: Additional keyword arguments
        :return: the annotated YAML object
        """
        data: object = annotate(newc(name, args), node)
        yield data
        base_ctor: "_CtorFuncType" = cast(
            "_CtorFuncType", getattr(base_cls, name)
        )
        generator: Iterator[object] = cast(
            Iterator[object], base_ctor(inst, node, *args, **kwargs)
        )
        gdata: object = {}
        for item in generator:
            gdata = item
        merge_data(data, gdata)

    return gctor


def annotate_constructed_objects(*spec: "_CtorSpecType") -> "_CtorDecorType":
    """
    Annotate constructed YAML objects with their location.

    :param spec: Each argument is a triple holding constructor name, tag, and
        a Boolean value, respectively, saying whether the constructor is a
        generator or not. Constructor name specifies the constructor involved
        in annotating YAML object during its construction, tag is a tag
        associated with the constructor or `None`
    :return: the decorator function that patches the class
    """

    def patch_constructors(cls: "type[_CtorType]") -> "type[_CtorType]":
        """
        Patch YAML object constructors.

        :param cls: The constructor class
        :return: the patched constructor class

        A decorator that patches YAML object constructors of the given class
        with the ability to annotate constructed objects with their location.
        """
        base_cls: "type[_CtorType]" = cls.__bases__[0]
        for name, tag, is_gen in spec:
            mfunc: "_CtorFuncType" = (
                make_gctor(name, base_cls)
                if is_gen
                else make_ctor(name, base_cls)
            )
            setattr(cls, name, mfunc)
            if tag is not None:
                cast("dict[str, _CtorFuncType]", cls.yaml_constructors)[
                    tag
                ] = mfunc
        return cls

    return patch_constructors


def getloc(obj: YamlDataType) -> Location:
    """
    Get the location of the object.

    :param obj: The object
    :return: the location of the object
    """
    return cast(
        Annotation, getattr(obj, ANNOTATION_SLOT, Annotation())
    ).location


def keyloc(obj: DictType, kobj: object) -> Location:
    """
    Get the location of the key object.

    :param obj: The annotated `dict` object
    :param kobj: The key
    :return: the location of the key
    """
    if not hasattr(obj, KEYLOC_SLOT):
        key2loc: "dict[YamlDataType, Location]" = {}
        for key in obj:
            key2loc[cast(YamlDataType, key)] = getloc(cast(YamlDataType, key))
        setattr(obj, KEYLOC_SLOT, key2loc)
    return cast("dict[object, Location]", getattr(obj, KEYLOC_SLOT)).get(
        kobj, Location()
    )


def is_null(obj: YamlDataType) -> bool:
    """
    Test whether the object is null.

    :param obj: The annotated object
    :return: `True` if *obj* is null
    """
    return isinstance(obj, NullType)


def is_bool(obj: YamlDataType) -> bool:
    """
    Test whether the object has Boolean type.

    :param obj: The annotated object
    :return: `True` if *obj* has Boolean type
    """
    return isinstance(obj, BoolType)



#
# File:    ./src/vutils/yaml/__init__.pyi
# Author:  Jiří Kučera <sanczes AT gmail.com>
# Date:    2022-06-23 10:01:39 +0200
# Project: vutils-yaml: Working with YAML format
#
# SPDX-License-Identifier: MIT
#

from collections.abc import Callable
from typing import Protocol, TypeVar

import yaml
from _typeshed import SupportsRead
from typing_extensions import TypeAlias

_T = TypeVar("_T")

PyList: TypeAlias = list[object]
PySet: TypeAlias = set[object]
PyDict: TypeAlias = dict[object, object]

_MarkType: TypeAlias = yaml.error.Mark
_NodeType: TypeAlias = yaml.nodes.Node
_CtorType: TypeAlias = yaml.constructor.BaseConstructor
_CtorSpecType: TypeAlias = tuple[str, str | None, bool]
_CtorDecorType: TypeAlias = Callable[[type[_CtorType]], type[_CtorType]]
_StreamType: TypeAlias = str | bytes | SupportsRead[str] | SupportsRead[bytes]

def _new_date(cls: type[_T], *args: object) -> _T: ...
def _new_datetime(cls: type[_T], *args: object, **kwargs: object) -> _T: ...

class _CtorFuncType(Protocol):
    def __call__(
        self, inst: _CtorType, node: _NodeType, *args: object, **kwargs: object
    ) -> object: ...



#
# File:    ./tests/unit/common.py
# Author:  Jiří Kučera <sanczes AT gmail.com>
# Date:    2022-07-07 18:20:52 +0200
# Project: vutils-yaml: Working with YAML format
#
# SPDX-License-Identifier: MIT
#
"""Shared test code and data."""

from vutils.testing.utils import ClassLikeSymbol


class PySet(metaclass=ClassLikeSymbol):
    """PySet class."""

    __slots__ = ()


class PyList(metaclass=ClassLikeSymbol):
    """PyList class."""

    __slots__ = ()


class PyDict(metaclass=ClassLikeSymbol):
    """PyDict class."""

    __slots__ = ()


SYMBOLS = (
    "_T",
    PySet,
    PyList,
    PyDict,
    "_MarkType",
    "_NodeType",
    "_CtorType",
    "_CtorSpecType",
    "_CtorDecorType",
    "_StreamType",
    "_new_date",
    "_new_datetime",
    "_CtorFuncType",
)
SLOT = "__foo__"
VALUE = "bar:42"



#
# File:    ./tests/unit/test_utils.py
# Author:  Jiří Kučera <sanczes AT gmail.com>
# Date:    2022-07-07 15:36:27 +0200
# Project: vutils-yaml: Working with YAML format
#
# SPDX-License-Identifier: MIT
#
"""Test `vutils.yaml.utils` module."""

import datetime

import yaml
from vutils.testing.mock import make_mock
from vutils.testing.testcase import TestCase
from vutils.testing.utils import make_type
from vutils.validator.value import Location

from vutils.yaml.load import load_yaml
from vutils.yaml.utils import (
    Annotation,
    BoolType,
    BytesType,
    DateTimeType,
    DateType,
    DictType,
    FloatType,
    IntType,
    ListType,
    NullType,
    SetType,
    StrType,
    annotate,
    getloc,
    is_bool,
    is_null,
    keyloc,
    newc,
    obj2xobj,
)

from .common import SLOT, VALUE


class AnnotationTestCase(TestCase):
    """Test case for `Annotation`."""

    __slots__ = ()

    def test_annotation_initialization(self):
        """Test `Annotation` initialization."""
        location = Location("./foo.yml", 1, 2)

        self.assertEqual(f"{Annotation().location}", "")
        self.assertIs(Annotation(location).location, location)


class NullTypeTestCase(TestCase):
    """Test case for `NullType`."""

    __slots__ = ()

    def test_null_type_behaves_like_false(self):
        """Test whether `NullType` object behaves like `False`."""
        self.assertFalse(NullType())

    def test_hashing_and_equality(self):
        """Test `__hash__` and `__eq__`."""
        self.assertEqual(hash(NullType()), hash(None))

        self.assertEqual(NullType(), None)
        self.assertEqual(NullType(), NullType())
        self.assertNotEqual(NullType(), 0)


class BoolTypeTestCase(TestCase):
    """Test case for `BoolType`."""

    __slots__ = ()

    def test_bool_type_can_hold_value(self):
        """Test whether a `BoolType` object can hold a Boolean value."""
        self.assertTrue(BoolType(True))
        self.assertFalse(BoolType(False))

    def test_hashing_and_equality(self):
        """Test `__hash__` and `__eq__`."""
        self.assertEqual(hash(BoolType(False)), hash(False))
        self.assertEqual(hash(BoolType(True)), hash(True))

        self.assertEqual(BoolType(False), BoolType(False))
        self.assertEqual(BoolType(False), False)
        self.assertEqual(BoolType(True), BoolType(True))
        self.assertEqual(BoolType(True), True)

        self.assertNotEqual(BoolType(False), BoolType(True))
        self.assertNotEqual(BoolType(False), True)
        self.assertNotEqual(BoolType(True), BoolType(False))
        self.assertNotEqual(BoolType(True), False)

        self.assertEqual(BoolType(False), 0)
        self.assertEqual(BoolType(False), IntType(0))
        self.assertEqual(BoolType(True), 1)
        self.assertEqual(BoolType(True), IntType(1))
        self.assertNotEqual(BoolType(True), 2)

        self.assertEqual(BoolType(False), 0.0)
        self.assertEqual(BoolType(False), FloatType(0.0))
        self.assertEqual(BoolType(True), 1.0)
        self.assertEqual(BoolType(True), FloatType(1.0))
        self.assertNotEqual(BoolType(True), 2.0)

        self.assertNotEqual(BoolType(False), [])
        self.assertNotEqual(BoolType(True), [])


class YamlDataTypeTestCase(TestCase):
    """Test case for `YamlDataType` and its subclasses."""

    __slots__ = ()

    def do_arithmetic_check(self, cls, zero, unit):
        """
        Test arithmetic types.

        :param cls: The arithmetic type
        :param zero: The zero element
        :param unit: The unit element
        """
        self.assertEqual(cls(unit), unit)
        self.assertEqual(cls(unit) + cls(unit), unit + unit)
        self.assertLess(cls(unit), unit + unit)
        self.assertFalse(cls(zero))
        self.assertTrue(cls(unit))

    def do_sequence_check(self, cls, sequence, empty):
        """
        Test sequence types.

        :param cls: The sequence type
        :param sequence: The nonempty sequence
        :param empty: The empty sequence
        """
        self.assertEqual(cls(sequence), sequence)
        self.assertEqual(cls(sequence)[0], sequence[0])
        self.assertTrue(cls(sequence))
        self.assertFalse(cls(empty))

    def do_extensibility_check(self, obj):
        """
        Test whether *obj* is extensible.

        :param obj: The YAML data object
        """
        self.assertFalse(hasattr(obj, SLOT))
        setattr(obj, SLOT, VALUE)
        self.assertTrue(hasattr(obj, SLOT))
        self.assertEqual(getattr(obj, SLOT), VALUE)

    @staticmethod
    def make_key(klass, key):
        """
        Make the key object.

        :param klass: The YAML data type
        :param key: The key
        :return: the *key* as instance of *klass*
        """
        return NullType() if klass is NullType else klass(key)

    def do_hashability_check(self, cls, key):
        """
        Test whether *cls* objects are hashable.

        :param cls: The YAML data type
        :param key1: The key
        """
        key1 = self.make_key(cls, key)
        key2 = self.make_key(cls, key)
        val1 = 1
        val2 = 2
        dobj = DictType({})

        dobj[key1] = val1
        dobj[key2] = val2

        self.assertEqual(dobj[key1], val2)
        self.assertEqual(dobj[key1], dobj[key])
        self.assertEqual(dobj[key1], dobj[key2])
        self.assertEqual(len(dobj), 1)
        for kobj in dobj:
            self.assertIs(kobj, key1)


class IntTypeTestCase(YamlDataTypeTestCase):
    """Test case for `IntType`."""

    __slots__ = ()

    def test_int_type_mimics_int(self):
        """Test whether `IntType` mimics `int`."""
        self.do_arithmetic_check(IntType, 0, 1)

    def test_int_type_is_extensible(self):
        """Test whether `IntType` can be extended."""
        self.do_extensibility_check(IntType(42))


class FloatTypeTestCase(YamlDataTypeTestCase):
    """Test case for `FloatType`."""

    __slots__ = ()

    def test_float_type_mimics_float(self):
        """Test whether `FloatType` mimics `float`."""
        self.do_arithmetic_check(FloatType, 0.0, 1.0)

    def test_float_type_is_extensible(self):
        """Test whether `FloatType` can be extended."""
        self.do_extensibility_check(FloatType(0.5))


class StrTypeTestCase(YamlDataTypeTestCase):
    """Test case for `StrType`."""

    __slots__ = ()

    def test_str_type_mimics_str(self):
        """Test whether `StrType` mimics `str`."""
        self.do_sequence_check(StrType, "abc", "")

    def test_str_type_is_extensible(self):
        """Test whether `StrType` can be extended."""
        self.do_extensibility_check(StrType("abc"))


class BytesTypeTestCase(YamlDataTypeTestCase):
    """Test case for `BytesType`."""

    __slots__ = ()

    def test_bytes_type_mimics_bytes(self):
        """Test whether `BytesType` mimics `byte`."""
        self.do_sequence_check(BytesType, b"abc", b"")

    def test_bytes_type_is_extensible(self):
        """Test whether `BytesType` can be extended."""
        self.do_extensibility_check(BytesType(b"\0"))


class ListTypeTestCase(YamlDataTypeTestCase):
    """Test case for `ListType`."""

    __slots__ = ()

    def test_list_type_mimics_list(self):
        """Test whether `ListType` mimics `List`."""
        self.do_sequence_check(ListType, [1, 2, 3], [])

    def test_list_type_is_extensible(self):
        """Test whether `ListType` can be extended."""
        self.do_extensibility_check(ListType([1, 2, 3]))


class SetTypeTestCase(YamlDataTypeTestCase):
    """Test case for `SetType`."""

    __slots__ = ()

    def test_set_type_mimics_set(self):
        """Test whether `SetType` mimics `set`."""
        self.assertEqual(SetType({1, 2.5, "a"}), {"a", 1, 2.5})
        self.assertIn("a", SetType({"a"}))
        self.assertFalse(SetType(set()))
        self.assertTrue(SetType({2}))

    def test_set_type_is_extensible(self):
        """Test whether `SetType` can be extended."""
        self.do_extensibility_check(SetType({"a"}))


class DictTypeTestCase(YamlDataTypeTestCase):
    """Test case for `DictType`."""

    __slots__ = ()

    def test_dict_type_mimics_dict(self):
        """Test whether `DictType` mimics `dict`."""
        self.assertEqual(DictType({"a": 1}), {"a": 1})
        self.assertIn("a", DictType({"a": "b"}))
        self.assertEqual(DictType({"a": "b"})["a"], "b")
        self.assertFalse(DictType({}))
        self.assertTrue(DictType({"a": 1}))

    def test_dict_type_is_extensible(self):
        """Test whether `DictType` can be extended."""
        self.do_extensibility_check(DictType({"a": 3.14}))

    def test_null_type_is_hashable(self):
        """Test whether `NullType` is hashable."""
        self.do_hashability_check(NullType, None)

    def test_bool_type_is_hashable(self):
        """Test whether `BoolType` is hashable."""
        self.do_hashability_check(BoolType, True)
        self.do_hashability_check(BoolType, False)

    def test_int_type_is_hashable(self):
        """Test whether `IntType` is hashable."""
        self.do_hashability_check(IntType, 1)

    def test_float_type_is_hashable(self):
        """Test whether `FloatType` is hashable."""
        self.do_hashability_check(FloatType, 1.6)

    def test_str_type_is_hashable(self):
        """Test whether `StrType` is hashable."""
        self.do_hashability_check(StrType, "abc")

    def test_bytes_type_is_hashable(self):
        """Test whether `BytesType` is hashable."""
        self.do_hashability_check(BytesType, b"\xff")

    def test_date_type_is_hashable(self):
        """Test whether `DateType` is hashable."""
        self.do_hashability_check(DateType, datetime.date(2008, 1, 1))

    def test_datetime_type_is_hashable(self):
        """Test whether `DateTimeType` is hashable."""
        self.do_hashability_check(DateTimeType, datetime.datetime(2008, 1, 1))


class DateTypeTestCase(YamlDataTypeTestCase):
    """Test case for `DateType`."""

    __slots__ = ()

    def test_date_type_mimics_date(self):
        """Test whether `DateType` mimics `datetime.date`."""
        date_a = datetime.date(2008, 7, 8)
        date_b = DateType(2008, 7, 8)
        date_c = DateType(date_a)

        self.assertEqual(date_b, date_a)
        self.assertEqual(date_b, date_c)

    def test_date_type_is_extensible(self):
        """Test whether `DateType` can be extended."""
        self.do_extensibility_check(DateType(2008, 1, 1))


class DateTimeTypeTestCase(YamlDataTypeTestCase):
    """Test case for `DateTimeType`."""

    __slots__ = ()

    def test_datetime_type_mimics_datetime(self):
        """Test whether `DateTimeType` mimics `datetime.datetime`."""
        tzinfo = datetime.tzinfo(datetime.timedelta(hours=2, minutes=0))
        dt_a = datetime.datetime(2008, 7, 8, 13, 17, 21, 476, tzinfo, fold=1)
        dt_b = DateTimeType(2008, 7, 8, 13, 17, 21, 476, tzinfo, fold=1)
        dt_c = DateTimeType(dt_a)

        self.assertEqual(dt_b, dt_a)
        self.assertEqual(dt_b, dt_c)

    def test_datetime_type_is_extensible(self):
        """Test whether `DateTimeType` can be extended."""
        self.do_extensibility_check(
            DateTimeType(
                2008,
                1,
                1,
                13,
                0,
                0,
                0,
                datetime.tzinfo(datetime.timedelta(hours=1)),
                fold=0,
            )
        )


class Obj2XObjTestCase(TestCase):
    """Test case for `obj2xobj`."""

    __slots__ = ()

    def do_test_obj2xobj(self, klass, obj):
        """
        Do the `obj2xobj` test.

        :param klass: The subclass of `YamlDataType`
        :param obj: The object
        """
        xobj = obj2xobj(obj)
        self.assertIsInstance(xobj, klass)
        self.assertEqual(xobj, obj)

    def test_obj2xobj(self):
        """Test `obj2xobj`."""
        self.do_test_obj2xobj(NullType, None)
        self.do_test_obj2xobj(BoolType, False)
        self.do_test_obj2xobj(IntType, 1)
        self.do_test_obj2xobj(FloatType, 1.25)
        self.do_test_obj2xobj(StrType, "abc")
        self.do_test_obj2xobj(BytesType, b"\12")
        self.do_test_obj2xobj(ListType, [1, 2])
        self.do_test_obj2xobj(SetType, {1, 2})
        self.do_test_obj2xobj(DictType, {1: 2})
        self.do_test_obj2xobj(DateType, datetime.date(2011, 4, 24))
        self.do_test_obj2xobj(DateTimeType, datetime.datetime(2009, 12, 26))
        with self.assertRaises(TypeError) as ctx:
            obj2xobj(obj2xobj)
        self.assertEqual(
            ctx.exception.args[0],
            "<class 'function'> object cannot be made extensible",
        )


class NewCTestCase(TestCase):
    """Test case for `newc`."""

    __slots__ = ()

    def do_test_newc(self, ctor, obj, *args):
        """
        Do the `newc` test.

        :param ctor: The name of the constructor
        :param obj: The object to be used for comparison
        :param args: Additional arguments to `newc`
        """
        cobj = newc(ctor, args)
        self.assertIs(type(cobj), type(obj))
        self.assertEqual(cobj, obj)

    def test_newc(self):
        """Test `newc`."""

        def eqfn(inst, other):
            """
            Equality test.

            :param inst: The instance of the owner object
            :param other: The instance of the other object
            :return: `True` if the two objects are equal
            """
            return isinstance(other, type(inst)) and inst.foo == other.foo

        tcls = make_type("TestType", members={"foo": 42, "__eq__": eqfn})

        self.do_test_newc("construct_yaml_set", set())
        self.do_test_newc("construct_yaml_map", {})
        self.do_test_newc("construct_yaml_object", tcls(), tcls)
        self.do_test_newc("construct_yaml_seq", [])


class AnnotateTestCase(TestCase):
    """Test case for `annotate`."""

    __slots__ = ("mark", "node")

    def setUp(self):
        """Set up the test."""
        mark = make_mock()
        mark.name = "./foo.yml"
        mark.line = 41
        mark.column = 7

        node = make_mock()
        node.start_mark = mark

        self.mark = mark
        self.node = node

    def test_annotate(self):
        """Test `annotate`."""
        obj = annotate({}, self.node)
        loc = getloc(obj)

        self.assertEqual(loc.path, self.mark.name)
        self.assertEqual(loc.line, self.mark.line + 1)
        self.assertEqual(loc.column, self.mark.column + 1)

        with self.assertRaises(yaml.constructor.ConstructorError) as ctx:
            annotate(loc, self.node)
        self.assertEqual(
            str(ctx.exception),
            (
                f"{loc.path}:{loc.line}:{loc.column}: "
                "<class 'vutils.validator.value.Location'>"
                " object cannot be made extensible"
            ),
        )


class GetLocTestCase(TestCase):
    """Test case for `getloc`."""

    __slots__ = ()

    def test_getloc(self):
        """Test `getloc`."""
        data = load_yaml("a: b")
        loc = getloc(data)

        self.assertEqual(loc.path, "<unicode string>")
        self.assertEqual(loc.line, 1)
        self.assertEqual(loc.column, 1)


class KeyLocTestCase(TestCase):
    """Test case for `keyloc`."""

    __slots__ = ()

    def test_keyloc(self):
        """Test `keyloc`."""
        data = load_yaml("{a: 1, b: 2, a: 3}")

        self.assertEqual(keyloc(data, "a").column, 2)
        self.assertEqual(keyloc(data, "b").column, 8)


class IsXTestCase(TestCase):
    """Test case for `is_null` and `is_bool`."""

    __slots__ = ()

    def test_isx(self):
        """Test `is_null` and `is_bool`."""
        self.assertTrue(is_null(NullType()))
        self.assertFalse(is_null(BoolType(False)))

        self.assertTrue(is_bool(BoolType(True)))
        self.assertFalse(is_bool(IntType(0)))



#
# File:    ./src/vutils/yaml/load.py
# Author:  Jiří Kučera <sanczes AT gmail.com>
# Date:    2022-06-24 00:41:59 +0200
# Project: vutils-yaml: Working with YAML format
#
# SPDX-License-Identifier: MIT
#
"""Load YAML document."""

from typing import TYPE_CHECKING, cast

import yaml

from vutils.yaml.utils import YamlDataType, annotate_constructed_objects

if TYPE_CHECKING:
    from collections.abc import Callable

    from vutils.yaml import _StreamType


@annotate_constructed_objects(
    ("construct_scalar", None, False),
    ("construct_sequence", None, False),
    ("construct_mapping", None, False),
    ("construct_pairs", None, False),
    ("construct_yaml_null", "tag:yaml.org,2002:null", False),
    ("construct_yaml_bool", "tag:yaml.org,2002:bool", False),
    ("construct_yaml_int", "tag:yaml.org,2002:int", False),
    ("construct_yaml_float", "tag:yaml.org,2002:float", False),
    ("construct_yaml_binary", "tag:yaml.org,2002:binary", False),
    ("construct_yaml_timestamp", "tag:yaml.org,2002:timestamp", False),
    ("construct_yaml_omap", "tag:yaml.org,2002:omap", True),
    ("construct_yaml_pairs", "tag:yaml.org,2002:pairs", True),
    ("construct_yaml_set", "tag:yaml.org,2002:set", True),
    ("construct_yaml_str", "tag:yaml.org,2002:str", False),
    ("construct_yaml_seq", "tag:yaml.org,2002:seq", True),
    ("construct_yaml_map", "tag:yaml.org,2002:map", True),
    ("construct_yaml_object", None, True),
)
class AnnotateConstructor(yaml.constructor.SafeConstructor):
    """Construct annotated YAML objects."""

    __slots__ = ()


class AnnotateLoader(
    yaml.reader.Reader,
    yaml.scanner.Scanner,
    yaml.parser.Parser,
    yaml.composer.Composer,
    AnnotateConstructor,
    yaml.resolver.Resolver,
):
    """Load YAML, annotate constructed objects."""

    __slots__ = ()

    def __init__(self, stream: "_StreamType") -> None:
        """
        Initialize the YAML loader.

        :param stream: The stream
        """
        yaml.reader.Reader.__init__(self, stream)
        yaml.scanner.Scanner.__init__(self)
        yaml.parser.Parser.__init__(self)
        yaml.composer.Composer.__init__(self)
        AnnotateConstructor.__init__(self)
        yaml.resolver.Resolver.__init__(self)


def load_yaml(stream: "_StreamType") -> YamlDataType:
    """
    Load YAML from the stream.

    :param stream: The stream
    """
    loader: AnnotateLoader = AnnotateLoader(stream)
    try:
        return cast(YamlDataType, loader.get_single_data())
    finally:
        cast("Callable[[], None]", loader.dispose)()



#
# File:    ./tests/unit/test_version.py
# Author:  Jiří Kučera <sanczes AT gmail.com>
# Date:    2022-06-23 10:01:39 +0200
# Project: vutils-yaml: Working with YAML format
#
# SPDX-License-Identifier: MIT
#
"""Test `vutils.yaml.version` module."""

from vutils.testing.testcase import TestCase

from vutils.yaml.version import __version__


class VersionTestCase(TestCase):
    """Test case for version."""

    __slots__ = ()

    def test_version(self):
        """Test if version is defined properly."""
        self.assertIsInstance(__version__, str)



#
# File:    ./src/vutils/yaml/__init__.py
# Author:  Jiří Kučera <sanczes AT gmail.com>
# Date:    2022-06-23 10:01:39 +0200
# Project: vutils-yaml: Working with YAML format
#
# SPDX-License-Identifier: MIT
#
"""`vutils.yaml` package."""



#
# File:    ./tests/unit/test_coverage.py
# Author:  Jiří Kučera <sanczes AT gmail.com>
# Date:    2022-07-07 22:39:20 +0200
# Project: vutils-yaml: Working with YAML format
#
# SPDX-License-Identifier: MIT
#
"""Coverage tests."""

import pytest
from vutils.testing.utils import cover_typing

from .common import SYMBOLS


@pytest.mark.order("last")
def test_typing_code_is_covered():
    """
    Ensure typing code coverage.

    This is a dummy test that executes `cover_typing` to ensure that the code
    under ``if TYPE_CHECKING:`` branch is covered. Since `cover_typing` reloads
    the module, run this test as last to prevent mangling of yet imported
    modules.
    """
    cover_typing("vutils.yaml.load", SYMBOLS)
    cover_typing("vutils.yaml.utils", SYMBOLS)



#
# File:    ./tests/unit/test_load.py
# Author:  Jiří Kučera <sanczes AT gmail.com>
# Date:    2022-07-07 15:36:46 +0200
# Project: vutils-yaml: Working with YAML format
#
# SPDX-License-Identifier: MIT
#
"""Test `vutils.yaml.load` module."""

from vutils.testing.testcase import TestCase

from vutils.yaml.load import load_yaml
from vutils.yaml.utils import (
    BoolType,
    BytesType,
    DateTimeType,
    DateType,
    DictType,
    FloatType,
    IntType,
    ListType,
    NullType,
    SetType,
    StrType,
    getloc,
)

YAML_STREAM = r"""
---
- null
- yes
- 1
- 3.14
- abc
- !!binary AA==
- []
- !!set
    ? a
    ? b
    ? c
- {}
- 2008-07-07
- 2008-07-07 13:15:22.489 +02:00
- !!omap
    - a: 1
    - b: 2
    - c: 3
- !!pairs
    - a: 1
    - b: 2
    - c: 3
"""
YAML_STREAM_NAME = "<unicode string>"


class LoadYamlTestCase(TestCase):
    """Test case for `load_yaml`."""

    __slots__ = ()

    def check_location(self, obj, name, line, column):
        """
        Check whether the object location matches the expectation.

        :param obj: The object
        :param name: The expected location name
        :param line: The expected location line
        :param column: The expected location column
        """
        loc = getloc(obj)
        self.assertEqual(loc.path, name)
        self.assertEqual(loc.line, line)
        self.assertEqual(loc.column, column)

    def do_test_data_are_annotated(self, obj, klass, line, column=3):
        """
        Check the data annotation.

        :param obj: The data object
        :param klass: The data type class
        :param line: The expected line of the data location
        :param column: The expected column of the data location
        """
        self.assertIsInstance(obj, klass)
        self.check_location(obj, YAML_STREAM_NAME, line, column)

    def test_loaded_data_are_annotated(self):
        """Test whether loaded data are properly annotated."""
        data = load_yaml(YAML_STREAM)

        self.do_test_data_are_annotated(data[0], NullType, 3)
        self.do_test_data_are_annotated(data[1], BoolType, 4)
        self.do_test_data_are_annotated(data[2], IntType, 5)
        self.do_test_data_are_annotated(data[3], FloatType, 6)
        self.do_test_data_are_annotated(data[4], StrType, 7)
        self.do_test_data_are_annotated(data[5], BytesType, 8)
        self.do_test_data_are_annotated(data[6], ListType, 9)
        self.do_test_data_are_annotated(data[7], SetType, 10)
        self.do_test_data_are_annotated(data[8], DictType, 14)
        self.do_test_data_are_annotated(data[9], DateType, 15)
        self.do_test_data_are_annotated(data[10], DateTimeType, 16)
        self.do_test_data_are_annotated(data[11], ListType, 17)
        self.do_test_data_are_annotated(data[12], ListType, 21)



#!/usr/bin/python3
#
# File:    ./setup.py
# Author:  Jiří Kučera <sanczes AT gmail.com>
# Date:    2022-06-23 10:01:39 +0200
# Project: vutils-yaml: Working with YAML format
#
# SPDX-License-Identifier: MIT
#
"""Setup for vutils-yaml."""

from setuptools import setup

setup()
